/*	
 *	clipdmv.c
 *
 *	P/ECE DMG-Video Emulator
 *
 *	CLiP - Common Library for P/ECE
 *	Copyright (C) 2001-2005 Naoyuki Sawa
 *
 *	* Fri Feb 18 20:18:00 JST 2005 Naoyuki Sawa
 *	- 쐬JnB
 */
#include "clip.h"

/****************************************************************************
 *
 ****************************************************************************/

DMGVIDEO dmgvideo;

/****************************************************************************
 *
 ****************************************************************************/

void
dmgvideo_reset(unsigned char* mem)
{
	DMGVIDEO* video = &dmgvideo;

	memset(video, 0, sizeof(DMGVIDEO));
	video->mem = mem;
}

int
dmgvideo_read(int addr)
{
	DMGVIDEO* video    = &dmgvideo;
	unsigned char* mem = video->mem;

	switch(addr) {
	//case 0xff40: /* $FF40 (LCDC) */
	//case 0xff41: /* $FF41 (STAT) */
	//case 0xff42: /* $FF42 (SCY) */
	//case 0xff43: /* $FF43 (SCX) */
	//case 0xff44: /* $FF44 (LY) */
	//case 0xff45: /* $FF45 (LYC) */
	//case 0xff46: /* $FF46 (DMA) */
	//case 0xff47: /* $FF47 (BGP) */
	//case 0xff48: /* $FF48 (OBP0) */
	//case 0xff49: /* $FF49 (OBP1) */
	//case 0xff4a: /* $FF4A (WY) */
	//case 0xff4b: /* $FF4B (WX) */
	}

	return mem[addr];
}

void
dmgvideo_write(int addr, int data)
{
	DMGVIDEO* video    = &dmgvideo;
	unsigned char* mem = video->mem;
	//
	unsigned char* clut;

	if(mem[addr] == data) {
		return;
	}
	mem[addr] = data;

	switch(addr) {
	//case 0xff40: /* $FF40 (LCDC) */
	//case 0xff41: /* $FF41 (STAT) */
	//case 0xff42: /* $FF42 (SCY) */
	//case 0xff43: /* $FF43 (SCX) */
	//case 0xff44: /* $FF44 (LY) */
	//case 0xff45: /* $FF45 (LYC) */

	case 0xff46: /* $FF46 (DMA) */
		memcpy(&mem[0xfe00], &mem[data<<8], 0xa0);
		mem[addr] = 0; /* 񏑂݌ô߁AL蓾ȂlɂĂ܂ */
		return;

	case 0xff47: /* $FF47 (BGP) */
	case 0xff48: /* $FF48 (OBP0) */
	case 0xff49: /* $FF49 (OBP1) */
		clut = &video->clut[(addr - 0xff47) * 4];
		*clut++ = data & 3; data >>= 2;
		*clut++ = data & 3; data >>= 2;
		*clut++ = data & 3; data >>= 2;
		*clut++ = data & 3;
		return;

	//case 0xff4a: /* $FF4A (WY) */
	//case 0xff4b: /* $FF4B (WX) */
	}
}

void
dmgvideo_update(unsigned char vbuff[/*160*144]*/])
{
	DMGVIDEO* video    = &dmgvideo;
	unsigned char* mem = video->mem;
	//
	int mode;

	/*{{*/
	if(!vbuff) return;
	memset(vbuff, 0, 160 * 144);
	video->vbuff      = vbuff;
	video->scan_start = 0;
	video->scan_end   = 144;
	/*}}*/

	mode = mem[0xff40]; /* $FF40 (LCDC) */
	if(mode & (1<<7)) {
		if(mode & (1<<1)) {
			dmgvideo_draw_sprite(1);
		}
		if(mode & (1<<0)) {
			dmgvideo_draw_bg();
			dmgvideo_draw_window();
		}
		if(mode & (1<<1)) {
			dmgvideo_draw_sprite(0);
		}
	}
}

void
dmgvideo_draw_bg()
{
	DMGVIDEO* video    = &dmgvideo;
	unsigned char* mem = video->mem;
	int scan_start     = video->scan_start;
	int scan_end       = video->scan_end;
	//
	int i;
	int x;
	int y;
	int col;
	int row;
	int code;
	int mode;
	int tile_map;

	mode = mem[0xff40]; /* $FF40 (LCDC) */
	tile_map = !(mode & (1<<3)) ? 0x9800 : 0x9c00;

	x = mem[0xff43]; /* $FF43 (SCX) */
	y = mem[0xff42]; /* $FF42 (SCY) */
	col = x >> 3; /* x = 0..255 -> col = 0..31 */
	row = y >> 3; /* y = 0..255 -> row = 0..31 */
	x = -(x & 7); /* x = -7..0 */
	y = -(y & 7); /* y = -7..0 */

	do {
		/* `̈̏ɊSɊORow̓XLbv܂B */
		if(y > scan_start - 8) {
			i = 160 / 8 + 1;
			do {
				code = mem[tile_map | row<<5 | col];
				if(!(mode & (1<<4))) {
					code = 0x100 + (char)code;
				}
				dmgvideo_draw_chr(x, y, code, 0/*BGP*/);
				//
				x += 8;
				col = (col + 1) & 31;
			} while(--i);
			x -= 8 * (160 / 8 + 1);
			col = (col - (160 / 8 + 1)) & 31;
		}
		y += 8;
		row = (row + 1) & 31;
	} while(y < scan_end); /* `̈֊SɊOꂽA܂B */
}

void
dmgvideo_draw_window()
{
	/* TODO: */
}

void
dmgvideo_draw_sprite(int priority)
{
	DMGVIDEO* video    = &dmgvideo;
	unsigned char* mem = video->mem;
	//
	int mode;

	mode = mem[0xff40]; /* $FF40 (LCDC) */
	if(!(mode & (1<<2))) {
		dmgvideo_draw_sprite_8x8(priority);
	} else {
		dmgvideo_draw_sprite_8x16(priority);
	}
}

void
dmgvideo_draw_sprite_8x8(int priority)
{
	DMGVIDEO* video    = &dmgvideo;
	unsigned char* mem = video->mem;
	//
	int i;
	int xpos;
	int ypos;
	int code;
	int attr;
	int palno;
	unsigned int* p;

	p = (unsigned int*)&mem[0xfea0];
	i = 40;
	do {
		attr = *--p;
		if((!priority && attr >= 0) ||
		   ( priority && attr <  0)) {

			ypos = (unsigned char)attr;
			attr >>= 8;
			xpos = (unsigned char)attr;
			attr >>= 8;
			code = (unsigned char)attr;
			attr >>= 8;

			ypos -= 16;
			xpos -= 8;
			palno = ((attr >> 4) & 1) + 1/*OBPx*/;

			switch((attr >> 5) & 3) {
			default: dmgvideo_draw_chr      (xpos, ypos, code, palno); break;
			case 1:  dmgvideo_draw_chr_revx (xpos, ypos, code, palno); break;
			case 2:  dmgvideo_draw_chr_revy (xpos, ypos, code, palno); break;
			case 3:  dmgvideo_draw_chr_revxy(xpos, ypos, code, palno); break;
			}
		}
	} while(--i);
}

void
dmgvideo_draw_sprite_8x16(int priority)
{
	DMGVIDEO* video    = &dmgvideo;
	unsigned char* mem = video->mem;
	//
	int i;
	int xpos;
	int ypos;
	int code;
	int attr;
	int palno;
	unsigned int* p;

	p = (unsigned int*)&mem[0xfea0];
	i = 40;
	do {
		attr = *--p;
		if((!priority && attr >= 0) ||
		   ( priority && attr <  0)) {

			ypos = (unsigned char)attr;
			attr >>= 8;
			xpos = (unsigned char)attr;
			attr >>= 8;
			code = (unsigned char)attr;
			attr >>= 8;

			ypos -= 16;
			xpos -= 8;
			code &= ~1;
			palno = ((attr >> 4) & 1) + 1/*OBPx*/;

			switch((attr >> 5) & 3) {
			default: dmgvideo_draw_chr      (xpos, ypos    , code + 0, palno);
			         dmgvideo_draw_chr      (xpos, ypos + 8, code + 1, palno); break;
			case 1:  dmgvideo_draw_chr_revx (xpos, ypos    , code + 0, palno);
			         dmgvideo_draw_chr_revx (xpos, ypos + 8, code + 1, palno); break;
			case 2:  dmgvideo_draw_chr_revy (xpos, ypos    , code + 1, palno);
			         dmgvideo_draw_chr_revy (xpos, ypos + 8, code + 0, palno); break;
			case 3:  dmgvideo_draw_chr_revxy(xpos, ypos    , code + 1, palno);
			         dmgvideo_draw_chr_revxy(xpos, ypos + 8, code + 0, palno); break;
			}
		}
	} while(--i);
}

/*
 * (]  5sNZ -> ]  4sNZ) x 32
 * (] c18C   -> ] c11C  ) x  8
 *
 * 18C -> 11C̈kp^[́ÂƂłB
 *
 *			]Line#	]Line#
 *	i=1		-----------	-----------
 *	i=1->0->3	 0		 0
 *	i=3->2		 1+ 2		 1
 *	i=2->1		 3+ 4		 2
 *	i=1->0->3	 5		 3
 *	i=3->2		 6+ 7		 4
 *	i=2->1		 8+ 9		 5
 *	i=1->0->3	10		 6
 *	i=3->2		11+12		 7
 *	i=2->1		13+14		 8
 *	i=1->0->3	15		 9
 *	i=3->2		16+17		10
 */
void
dmgvideo_reduce(unsigned char dst[/*128*88*/], const unsigned char src[/*160*144*/])
{
	int i;
	int x;
	int y;
	int z;

	z = 8;
	do {
		y = 11;
		i = 1;
		do {
			x = 32;
			if(--i) {
				do {
					dst[0] = (src[0] + src[160]) >> 1;
					dst[1] = (src[1] + src[161]) >> 1;
					dst[2] = (src[2] + src[162]) >> 1;
					dst[3] = (src[3] + src[163] +
					          src[4] + src[164]) >> 2;
					dst += 4;
					src += 5;
				} while(--x);
				src += 160;
			} else {
				do {
					dst[0] =  src[0];
					dst[1] =  src[1];
					dst[2] =  src[2];
					dst[3] = (src[3] + src[4]) >> 1;
					dst += 4;
					src += 5;
				} while(--x);
				i = 3;
			}
		} while(--y);
	} while(--z);
}

/****************************************************************************
 *
 ****************************************************************************/

#define DMGVIDEO_DRAW_CHR_IMPLEMENT
//
#define DMGVIDEO_REVX 0
#define DMGVIDEO_REVY 0
void dmgvideo_draw_chr(int xpos, int ypos, int code, int palno)
#include "clipdmvi.h"
#undef DMGVIDEO_REVX
#undef DMGVIDEO_REVY
//
#define DMGVIDEO_REVX 1
#define DMGVIDEO_REVY 0
void dmgvideo_draw_chr_revx(int xpos, int ypos, int code, int palno)
#include "clipdmvi.h"
#undef DMGVIDEO_REVX
#undef DMGVIDEO_REVY
//
#define DMGVIDEO_REVX 0
#define DMGVIDEO_REVY 1
void dmgvideo_draw_chr_revy(int xpos, int ypos, int code, int palno)
#include "clipdmvi.h"
#undef DMGVIDEO_REVX
#undef DMGVIDEO_REVY
//
#define DMGVIDEO_REVX 1
#define DMGVIDEO_REVY 1
void dmgvideo_draw_chr_revxy(int xpos, int ypos, int code, int palno)
#include "clipdmvi.h"
#undef DMGVIDEO_REVX
#undef DMGVIDEO_REVY
//
#undef DMGVIDEO_DRAW_CHR_IMPLEMENT

/****************************************************************************
 *
 ****************************************************************************/

#if 0
WX^}bv̒^tH[B悭ĝŕۑ
if(addr >= 0xff00) {
	switch(addr) {
	/*----- Joypad -----*/
	case 0xff00: /* $FF00 (P1) */
	/*----- Serial -----*/
	case 0xff01: /* $FF01 (SB) */
	case 0xff02: /* $FF02 (SC) */
	/*----- Timer -----*/
	case 0xff04: /* $FF04 (DIV) */
	case 0xff05: /* $FF05 (TIMA) */
	case 0xff06: /* $FF06 (TMA) */
	case 0xff07: /* $FF07 (TAC) */
	/*----- Interrupt -----*/
	case 0xff0f: /* $FF0F (IF) */
	case 0xffff: /* $FFFF (IE) */
	/*----- Sound -----*/
	case 0xff10: /* $FF10 (NR10) */
	case 0xff11: /* $FF11 (NR11) */
	case 0xff12: /* $FF12 (NR12) */
	case 0xff13: /* $FF13 (NR13) */
	case 0xff14: /* $FF14 (NR14) */
	case 0xff16: /* $FF16 (NR21) */
	case 0xff17: /* $FF17 (NR22) */
	case 0xff18: /* $FF18 (NR23) */
	case 0xff19: /* $FF19 (NR24) */
	case 0xff1a: /* $FF1A (NR30) */
	case 0xff1b: /* $FF1B (NR31) */
	case 0xff1c: /* $FF1C (NR32) */
	case 0xff1d: /* $FF1D (NR33) */
	case 0xff1e: /* $FF1E (NR34) */
	case 0xff20: /* $FF20 (NR41) */
	case 0xff21: /* $FF21 (NR42) */
	case 0xff22: /* $FF22 (NR43) */
	case 0xff23: /* $FF23 (NR44) */
	case 0xff24: /* $FF24 (NR50) */
	case 0xff25: /* $FF25 (NR51) */
	case 0xff26: /* $FF26 (NR52) */
	/*----- LCDC -----*/
	case 0xff40: /* $FF40 (LCDC) */
	case 0xff41: /* $FF41 (STAT) */
	case 0xff42: /* $FF42 (SCY) */
	case 0xff43: /* $FF43 (SCX) */
	case 0xff44: /* $FF44 (LY) */
	case 0xff45: /* $FF45 (LYC) */
	case 0xff46: /* $FF46 (DMA) */
	case 0xff47: /* $FF47 (BGP) */
	case 0xff48: /* $FF48 (OBP0) */
	case 0xff49: /* $FF49 (OBP1) */
	case 0xff4a: /* $FF4A (WY) */
	case 0xff4b: /* $FF4B (WX) */
	}
}
#endif
